"""
trionyx.trionyx.layouts
~~~~~~~~~~~~~~~~~~~~~~~

:copyright: 2018 by Maikel Martens
:license: GPLv3
"""
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.utils.translation import ugettext_lazy as _
from django.utils import timezone
from rest_framework.authtoken.models import Token
from trionyx.layout import (
    Container, Row, Column10, Column2, Column12, Column6,
    Panel, DescriptionList, TableDescription, Img, Table, Html, HtmlTemplate
)
from trionyx.renderer import datetime_value_renderer
from trionyx.trionyx.models import AuditLogEntry, LogEntry
from trionyx.views import tabs
from .renderers import render_level


@tabs.register('trionyx.profile')
def account_overview(obj):
    """Create layout for user profile"""
    token, created = Token.objects.get_or_create(user=obj)
    return Container(
        Row(
            Column2(
                Panel(
                    _('Avatar'),
                    Img(src="{}{}".format(
                        settings.MEDIA_URL if obj.avatar else settings.STATIC_URL,
                        obj.avatar if obj.avatar else 'img/avatar.png'
                    )),
                    collapse=True,
                ),
            ),
            Column10(
                Column6(
                    Panel(
                        _('Account information'),
                        DescriptionList(
                            'email',
                            'first_name',
                            'last_name',
                        ),
                    )
                ),
                Column6(
                    Panel(
                        # TODO Add token reset button
                        _('Settings'),
                        DescriptionList(
                            {
                                'label': _('Language'),
                                'value': _(obj.get_language_display()),
                            },
                            'timezone',
                            {
                                'label': _('API Token'),
                                'value': token.key,
                            },
                        ),
                    )
                )
            ),
        )
    )


@tabs.register('trionyx.user')
def trionyx_user(obj):
    """Create layout for admin user"""
    from trionyx.trionyx.views import create_permission_jstree
    from django.contrib.auth.models import Permission
    permissions = []
    if obj.is_superuser:
        permissions = Permission.objects.all()
    else:
        permissions.extend(list(obj.user_permissions.all()))
        for group in obj.groups.all():
            permissions.extend(list(group.permissions.all()))

    token, created = Token.objects.get_or_create(user=obj)
    return Container(
        Row(
            Column2(
                Panel(
                    _('Avatar'),
                    Img(src="{}{}".format(
                        settings.MEDIA_URL if obj.avatar else settings.STATIC_URL,
                        obj.avatar if obj.avatar else 'img/avatar.png'
                    )),
                    collapse=True,
                ),
            ),
            Column10(
                Panel(
                    _('Account information'),
                    DescriptionList(
                        'email',
                        'first_name',
                        'last_name',
                        'created_at',
                        'last_login',
                        'last_online',
                        'is_active',
                        'is_superuser',
                        'groups',
                        {
                            'label': _('API Token'),
                            'value': token.key,
                        },
                    ),
                ),
                Panel(
                    _('All active permissions'),
                    HtmlTemplate(
                        template_name='trionyx/base/permissions.html',
                        context={
                            'permission_jstree': create_permission_jstree(permissions, disabled=True),
                        },
                        css_files=['plugins/jstree/themes/default/style.css'],
                        js_files=['plugins/jstree/jstree.min.js'],
                    ),
                ),
            ),
        ),
    )


@tabs.register('auth.group')
def auth_group(obj):
    """Create layout for permission group"""
    from trionyx.trionyx.views import create_permission_jstree
    return [
        Column12(
            Panel(
                _('Info'),
                TableDescription(
                    'name'
                )
            )
        ),
        Column12(
            Panel(
                _('Permissions'),
                HtmlTemplate(
                    template_name='trionyx/base/permissions.html',
                    context={
                        'permission_jstree': create_permission_jstree(obj.permissions.all(), disabled=True),
                    },
                    css_files=['plugins/jstree/themes/default/style.css'],
                    js_files=['plugins/jstree/jstree.min.js'],
                ),
            ),
        )
    ]


@tabs.register('trionyx.log')
def trionyx_log(obj):
    """Create log layout"""
    return Container(
        Row(
            Column6(
                Panel(
                    _('Log info'),
                    TableDescription(
                        {
                            'label': _('Level'),
                            'value': render_level(obj)
                        },
                        'message',
                        {
                            'label': _('Location'),
                            'value': '{}:{}'.format(obj.file_path, obj.file_line)
                        },
                        'last_event',
                        'log_count',
                    )
                ),
                Panel(
                    _('Backtrace'),
                    Html(
                        '<pre>{}</pre>'.format(obj.traceback) if obj.traceback
                        else '<div class="alert alert-info" style="margin: 0;border-radius: 0;">No traceback</div>'
                    )
                )
            ),
            Column6(
                Panel(
                    _('Last log entries'),
                    Table(
                        obj.entries.select_related('user').order_by('-id')[:25],
                        'log_time',
                        'user',
                        'path',
                        'user_agent',
                        object=LogEntry()
                    )
                )
            )
        )
    )


def progress_renderer(value, *args, **kwargs):
    """Render progressbar"""
    return """<div class="progress">
        <div
            class="progress-bar progress-bar-theme"
            role="progressbar"
            aria-valuemin="0"
            aria-valuemax="100"
            aria-valuenow="{progress}"
            style="width: {progress}%"
        >
            {progress}%
        </div>
    </div>""".format(
        progress=value,
    )


@tabs.register('trionyx.task')
def task(obj):
    """Render task layout"""
    return Container(
        Row(
            Column6(
                Panel(
                    _('General'),
                    TableDescription(
                        'description',
                        'status',
                        'user',
                        'started_at',
                        {
                            'field': 'execution_time',
                            'renderer': lambda value, **options: str(timezone.timedelta(seconds=value)),
                        },
                        'result',
                    )
                )
            ),
            Column6(
                Panel(
                    _('Process'),
                    TableDescription(
                        {
                            'field': 'progress',
                            'renderer': progress_renderer,
                        },
                        {
                            'field': 'progress_output',
                            'renderer': lambda value, **options: '<br/>'.join(value),
                        }
                    )
                )
            )
        )
    )


def auditlog(obj):
    """Create auditlog history layout"""
    content_type = ContentType.objects.get_for_model(obj)
    return [
        Column12(
            Panel(
                _('{action} on {date} by {user}').format(
                    action=auditlog.get_action_display(),
                    date=datetime_value_renderer(auditlog.created_at),
                    user=auditlog.user if auditlog.user else 'System'
                ),
                Table(
                    [[field, *changes] for field, changes in auditlog.changes.items()],
                    {
                        'label': _('Field'),
                        'width': '10%',
                    },
                    {
                        'label': _('Old value'),
                        'width': '45%',
                    },
                    {
                        'label': _('New value'),
                        'width': '45%',
                    },
                )
            )
        ) for auditlog in AuditLogEntry.objects.filter(content_type=content_type, object_id=obj.id).order_by('-created_at')
    ]



"""
trionyx.trionyx.search
~~~~~~~~~~~~~~~~~~~~~~

Add search and global search to models

:copyright: 2018 by Maikel Martens
:license: GPLv3
"""
from watson import search

from trionyx.config import models_config


class ModelSearchAdapter(search.SearchAdapter):
    """Generic search adapter for Trionyx models"""

    def get_title(self, obj):
        """Set search entry title for object"""
        from trionyx.renderer import LazyFieldRenderer
        search_title = self.get_model_config_value(obj, 'search_title')

        if not search_title:
            return super().get_title(obj)

        return search_title.format(**{
            field.name: LazyFieldRenderer(obj, field.name)
            for field in models_config.get_config(obj).get_fields(True, True)
        })

    def get_description(self, obj):
        """Set search entry description for object"""
        from trionyx.renderer import LazyFieldRenderer
        search_description = self.get_model_config_value(obj, 'search_description')

        if not search_description:
            return super().get_description(obj)

        return search_description.format(**{
            field.name: LazyFieldRenderer(obj, field.name)
            for field in models_config.get_config(obj).get_fields(True, True)
        })

    def get_model_config_value(self, obj, name):
        """Get config value for given model"""
        config = models_config.get_config(obj)
        return getattr(config, name)

    def get_url(self, obj):
        """Return the URL of the given obj."""
        return models_config.get_config(obj).get_absolute_url(obj)


def auto_register_search_models():
    """Auto register all search models"""
    for config in models_config.get_all_configs(False):
        if config.disable_search_index:
            continue

        search.register(
            config.model.objects.get_queryset(),
            ModelSearchAdapter,
            fields=config.search_fields,
            exclude=config.search_exclude_fields,
        )



"""
trionyx.data
~~~~~~~~~~~~

:copyright: 2019 by Maikel Martens
:license: GPLv3
"""
from .countries import COUNTRIES
from .currencies import CURRENCIES
from .timezones import TIMEZONES

__all__ = ['COUNTRIES', 'CURRENCIES', 'TIMEZONES']



# copyright 2019 by Maikel Martens
#
# license GPLv3

"""
Layout and Components
=====================

Layouts are used to render a view for an object.
Layouts are defined and registered in layouts.py in an app.


**Example of a tab layout for the user profile:**

.. code-block:: python

    @tabs.register('trionyx.profile')
    def account_overview(obj):
        return Container(
            Row(
                Column2(
                    Panel(
                        'Avatar',
                        Img(src="{}{}".format(settings.MEDIA_URL, obj.avatar)),
                        collapse=True,
                    ),
                ),
                Column10(
                    Panel(
                        'Account information',
                        DescriptionList(
                            'email',
                            'first_name',
                            'last_name',
                        ),
                    )
                ),
            )
        )


"""
import re

from django import template
from django.utils.functional import cached_property
from django.template.loader import render_to_string

from trionyx import utils

register = template.Library()


class Layout:
    """Layout object that holds components"""

    def __init__(self, *components, **options):
        """Initialize Layout"""
        self.object = False
        self.components = list(components)
        self.options = options

    def __getitem__(self, slice):
        """Get component item"""
        return self.components[slice]

    def __setitem__(self, slice, value):
        """Set component"""
        self.components[slice] = value

    def __delitem__(self, slice):
        """Delete component"""
        del self.components[slice]

    def __len__(self):
        """Get component length"""
        return len(self.components)

    def get_paths(self):
        """Get all paths in layout for easy lookup"""
        def generate_paths(component=None, path=None, index=None):
            """Generate all paths in layout"""
            path = [
                *path,
                str(component.__class__.__name__ + '-' + str(index)).lower()
            ]
            paths = {'.'.join(path): component}

            for index, comp in enumerate(component.components):
                paths.update(generate_paths(comp, path, index))
            return paths

        paths = {}
        for index, comp in enumerate(self.components):
            paths.update(generate_paths(comp, [], index))

        # Can't cache paths because after one change the are not valid
        return paths

    def find_component_by_path(self, path):
        """Find component by path, gives back component and parent"""
        path_re = re.compile(r'([\wd]+)(\[(\d+)\])?')
        new_path = []
        for p in path.lower().split('.'):
            match = path_re.match(p)
            new_path.append('{}-{}'.format(
                match.group(1),
                match.group(3) if match.group(3) else 0
            ))

        paths = self.get_paths()
        return (
            paths.get('.'.join(new_path)),
            paths.get('.'.join(new_path[:-1]))
        )

    def find_component_by_id(self, id=None, current_comp=None):
        """Find component by id, gives back component and parent"""
        current_comp = current_comp if current_comp else self

        for comp in current_comp.components:
            if id and comp.id == id:
                return (comp, current_comp)

        for comp in current_comp.components:
            r_comp, r_parent = self.find_component_by_id(id=id, current_comp=comp)
            if r_comp:
                return (r_comp, r_parent)

        return (None, None)

    def render(self, request=None):
        """Render layout for given request"""
        return render_to_string('trionyx/layout.html', {
            'layout': self,
            'css_files': self.collect_css_files(),
            'js_files': self.collect_js_files(),
        }, request)

    def collect_css_files(self, component=None):
        """Collect all css files"""
        component = component if component else self
        files = getattr(component, 'css_files', None)
        files = files if files else []

        for comp in component.components:
            files.extend(self.collect_css_files(comp))

        return list(set(files))

    def collect_js_files(self, component=None):
        """Collect all js files"""
        component = component if component else self
        files = getattr(component, 'js_files', None)
        files = files if files else []

        for comp in component.components:
            files.extend(self.collect_js_files(comp))

        return list(set(files))

    def set_object(self, object):
        """
        Set object for rendering layout and set object to all components

        :param object:
        :return:
        """
        if self.object is False:
            self.object = object

        # Pass object along to child components for rendering
        for component in self.components:
            component.set_object(self.object)

    def add_component(self, component, id=None, path=None, before=False):
        """
        Add component to existing layout can insert component before or after component

        :param component:
        :param id: component id
        :param path: component path, example: container.row.column6[1].panel
        :return:
        """
        if not id and not path:
            if before:
                self.components.append(component)
            else:
                self.components.insert(0, component)

        if id:
            comp, parent = self.find_component_by_id(id)
        else:
            comp, parent = self.find_component_by_path(path)

        if not comp:
            raise Exception('Could not add component: Unknown path {} or id {}'.format(path, id))

        if parent:
            index = parent.components.index(comp) if before else parent.components.index(comp) + 1
            parent.components.insert(index, component)
        elif comp:
            index = self.components.index(comp) if before else self.components.index(comp) + 1
            self.components.insert(index, component)

    def delete_component(self, id=None, path=None):
        """
        Delete component for given path or id

        :param id: component id
        :param path: component path, example: container.row.column6[1].panel
        :return:
        """
        if not id and not path:
            raise Exception('You must supply an id or path')

        if id:
            comp, parent = self.find_component_by_id(id)
        else:
            comp, parent = self.find_component_by_path(path)

        if not comp:
            raise Exception('Could not delete component: Unknown path {} or id {}'.format(path, id))

        if parent:
            parent.components.remove(comp)
        elif comp:
            self.components.remove(comp)


class Component:
    """Base component"""

    template_name = None
    """Component template to be rendered, default template only renders child components"""

    js_files = None
    """List of required javascript files"""

    css_files = None
    """List of required css files"""

    def __init__(self, *components, **options):
        """Initialize Component"""
        self.id = options.get('id')
        self.components = list(filter(None, components))
        self.parent = None
        self.object = options.get('object', False)
        self.context = {}
        self.request = None

        # set options on object
        for key, value in options.items():
            setattr(self, key, value)

    @cached_property
    def css_id(self):
        """Generate random css id for component"""
        return 'component-{}'.format(utils.random_string(6))

    def set_object(self, object):
        """
        Set object for rendering component and set object to all components

        :param object:
        :return:
        """
        if self.object is False:
            self.object = object
        else:
            object = self.object

        # Pass object along to child components for rendering
        for component in self.components:
            component.set_object(object)

    def render(self, context, request=None):
        """Render component"""
        context['component'] = self
        self.context = context
        self.request = request
        return render_to_string(self.template_name, context, request)


class ComponentFieldsMixin:
    """Mixin for adding fields support and rendering of object(s) with fields."""

    fields = []
    """
    List of fields to be rendered. Item can be a string or dict, default options:

    - **field**: Name of object attribute or dict key to be rendered
    - **label**: Label of field
    - **value**: Value to be rendered
    - **format**: String format for rendering field, default is '{0}'
    - **renderer**: Render function for rendering value, result will be given to format. (lambda value, **options: value)

    Based on the order the fields are in the list a __index__ is set with the list index,
    this is used for rendering a object that is a list.

    .. code-block:: python

        fields = [
            'first_name',
            'last_name'
        ]

        fields = [
            'first_name',
            {
                'label': 'Real last name',
                'value': object.last_name
            }
        ]
    """

    fields_options = {}
    """
    Options available for the field, this is not required to set options on field.

    - **default**: Default option value when not set.

    .. code-block:: python

        fields_options = {
            'width': {
                'default': '150px',
            }
        }

    """

    objects = []
    """
    List of object to be rendered, this can be a QuerySet, list or string.
    When its a string it will get the attribute of the object.

    The items in the objects list can be a mix of Models, dicts or lists.
    """

    def add_field(self, field, index=None):
        """Add field"""
        self.fields = list(self.fields)
        if index is not None:
            self.fields.insert(index, field)
        else:
            self.fields.append(field)

    def get_fields(self):
        """Get all fields"""
        if not hasattr(self, '__fields'):
            self.__fields = [
                self.parse_field(field, index)
                for index, field in enumerate(getattr(self, 'fields', []))
                if not (field is False or field is None)
            ]
        return self.__fields

    def parse_field(self, field_data, index=0):
        """Parse field and add missing options"""
        field = {
            '__index__': index,
        }

        if isinstance(field_data, str):
            field.update(self.parse_string_field(field_data))
        elif isinstance(field_data, dict):
            field.update(field_data)
        else:
            raise TypeError('Expected a str or dict get {}'.format(type(field_data)))

        if 'field' not in field:
            field['field'] = None

        if 'label' not in field and field['field']:
            try:
                field['label'] = self.object._meta.get_field(field['field']).verbose_name.capitalize()
            except Exception:
                field['label'] = field['field'].replace('_', '').capitalize()
        elif 'label' not in field:
            field['label'] = ''

        if 'format' not in field:
            field['format'] = '{0}'

        # Set default options
        for name, options in self.fields_options.items():
            if 'default' in options and name not in field:
                field[name] = options['default']

        return field

    def parse_string_field(self, field_data):
        """
        Parse a string field to dict with options

        String value is used as field name. Options can be given after = symbol.
        Where key value is separated by : and different options by ;, when no : is used then the value becomes True.

        **Example 1:** `field_name`

        .. code-block:: python

            # Output
            {
                'field': 'field_name'
            }

        **Example 3** `field_name=option1:some value;option2: other value`

        .. code-block:: python

            # Output
            {
                'field': 'field_name',
                'option1': 'some value',
                'option2': 'other value',
            }

        **Example 3** `field_name=option1;option2: other value`

        .. code-block:: python

            # Output
            {
                'field': 'field_name',
                'option1': True,
                'option2': 'other value',
            }

        :param str field_data:
        :return dict:
        """
        field_name, *data = field_data.split('=', 1)
        field = {
            'field': field_name,
        }

        for option_string in ''.join(data).split(';'):
            option, *value = option_string.split(':')
            if option.strip():
                field[option.strip()] = value[0].strip() if value else True

        return field

    def render_field(self, field, data):
        """Render field for given data"""
        from trionyx.renderer import renderer

        if 'value' in field:
            value = field['value']
        elif isinstance(data, object) and field['field'] and hasattr(data, field['field']):
            value = getattr(data, field['field'])
            if 'renderer' not in field:
                value = renderer.render_field(data, field['field'], **field)
        elif isinstance(data, dict) and field['field'] in data:
            value = data.get(field['field'])
        elif isinstance(data, list) and field['__index__'] < len(data):
            value = data[field['__index__']]
        else:
            return ''

        options = {key: value for key, value in field.items() if key not in ['value', 'data_object']}
        if 'renderer' in field:
            value = field['renderer'](value, data_object=data, **options)
        elif isinstance(value, Component):
            value.set_object(self.object)
            value = value.render(self.context.copy(), self.request)
        else:
            value = renderer.render_value(value, data_object=data, **options)

        return field['format'].format(value)

    def get_rendered_object(self, obj=None):
        """Render object"""
        obj = obj if obj else self.object
        return [
            {
                **field,
                'value': self.render_field(field, obj)
            }
            for field in self.get_fields()
        ]

    def get_rendered_objects(self):
        """Render objects"""
        objects = self.objects

        if isinstance(objects, str):
            objects = getattr(self.object, objects).all()

        return [
            self.get_rendered_object(obj)
            for obj in objects
        ]


# =============================================================================
# Simple HTML tags
# =============================================================================
class HtmlTemplate(Component):
    """HtmlTemplate render django html template"""

    def __init__(self, template_name, context=None, css_files=None, js_files=None):
        """Initialize HtmlTemplate"""
        super().__init__()
        self.template_name = template_name
        self.context = context
        self.css_files = css_files if css_files else []
        self.js_files = js_files if js_files else []

    def render(self, context, request=None):
        """Render component"""
        context['component'] = self
        context.update(self.context)
        self.context = context
        self.request = request
        return render_to_string(self.template_name, context, request)


class HtmlTagWrapper(Component):
    """HtmlTagWrapper wraps given component in given html tag"""

    template_name = 'trionyx/components/html_tag.html'

    tag = 'div'
    """Html tag nam"""

    attr = None
    """Dict with html attributes"""

    def __init__(self, *args, **kwargs):
        """Initialize HtmlTagWrapper"""
        super().__init__(*args, **kwargs)
        self.attr = self.attr.copy() if self.attr else {}

    def get_attr_text(self):
        """Get html attr text to render in template"""
        return ' '.join([
            '{}="{}"'.format(key, value)
            for key, value in self.attr.items()
        ])


class Html(HtmlTagWrapper):
    """Html single html tag"""

    template_name = 'trionyx/components/html.html'
    tag = None

    valid_attr = []
    """Valid attributes that can be used"""

    def __init__(self, html=None, **kwargs):
        """Init Html"""
        super().__init__(**kwargs)
        self.html = html
        for key, value in kwargs.items():
            if key in self.valid_attr:
                self.attr[key] = value


class Img(Html):
    """Img tag"""

    tag = 'img'
    valid_attr = ['src', 'width']
    attr = {
        'width': '100%',
    }


class Input(Html):
    """Input tag"""

    template_name = 'trionyx/components/input.html'

    valid_attr = ['name', 'value', 'type', 'placeholder', 'class']
    attr = {
        'type': 'text',
        'class': 'form-control',
    }

    def __init__(self, form_field=None, has_error=False, **kwargs):
        """Init input"""
        super().__init__(None, **kwargs)
        self.has_error = has_error

        if form_field:
            self.attr['name'] = form_field.name
            self.attr['value'] = form_field.value()
            self.has_error = form_field.errors

        if 'value' in self.attr and self.attr['value'] is None:
            self.attr['value'] = ''


class Button(Html):
    """
    Bootstrap button

    - link_url
    - dialog_url
    - onClick
    """

    tag = 'button'
    valid_attr = ['onClick', 'class']
    attr = {
        'class': 'btn btn-flat btn-default'
    }

    def __init__(self, label, link_url=None, dialog_url=None, dialog_options=None, **options):
        """Init button"""
        dialog_options = dialog_options if dialog_options else {}
        if not options.get('onClick') and (link_url or dialog_url):
            if link_url:
                options['onClick'] = "window.location.href='{}'; return false;".format(link_url)
            else:
                options['onClick'] = "openDialog('{}', {}); return false;".format(dialog_url, self.format_dialog_options(dialog_options))
        super().__init__(html=label, **options)

    def format_dialog_options(self, dialog_options):
        """Fromat options to JS dict"""
        return '{{ {} }}'.format(','.join(
            ("{}:{}" if key == 'callback' else "{}:'{}'").format(key, value)
            for key, value in dialog_options.items()))


# =============================================================================
# Bootstrap grid system
# =============================================================================
class Container(HtmlTagWrapper):
    """Bootstrap container"""

    attr = {
        'class': 'container-fluid'
    }


class Row(HtmlTagWrapper):
    """Bootstrap row"""

    attr = {
        'class': 'row'
    }


class Column(HtmlTagWrapper):
    """Bootstrap Column"""

    size = 'md'
    columns = 1

    def __init__(self, *args, **kwargs):
        """Initialize Column"""
        super().__init__(*args, **kwargs)
        self.attr['class'] = '-'.join(x for x in ['col', str(self.size), str(self.columns)] if x)


class Column2(Column):
    """Bootstrap Column 2"""

    columns = 2


class Column3(Column):
    """Bootstrap Column 3"""

    columns = 3


class Column4(Column):
    """Bootstrap Column 4"""

    columns = 4


class Column5(Column):
    """Bootstrap Column 5"""

    columns = 5


class Column6(Column):
    """Bootstrap Column 6"""

    columns = 6


class Column7(Column):
    """Bootstrap Column 7"""

    columns = 7


class Column8(Column):
    """Bootstrap Column 8"""

    columns = 8


class Column9(Column):
    """Bootstrap Column 9"""

    columns = 9


class Column10(Column):
    """Bootstrap Column 10"""

    columns = 10


class Column11(Column):
    """Bootstrap Column 11"""

    columns = 11


class Column12(Column):
    """Bootstrap Column 12"""

    columns = 12


# =============================================================================
# Bootstrap elements
# =============================================================================
class Panel(Component):
    """
    Bootstrap panel available options

    - title
    - footer_components
    - collapse
    - contextual: primary, success, info, warning, danger

    """

    template_name = 'trionyx/components/panel.html'

    collapse = True

    def __init__(self, title, *components, **options):
        """Init Panel"""
        super().__init__(*components, **options)
        self.title = title


class DescriptionList(Component, ComponentFieldsMixin):
    """
    Bootstrap description, fields are the params. available options

    - horizontal
    """

    template_name = 'trionyx/components/description_list.html'
    horizontal = True
    no_data_message = "There is no data"

    def __init__(self, *fields, **options):
        """Init panel"""
        super().__init__(**options)
        self.fields = fields


class TableDescription(Component, ComponentFieldsMixin):
    """Bootstrap table description, fields are the params"""

    template_name = 'trionyx/components/table_description.html'

    fields_options = {
        'width': {
            'default': '150px',
        }
    }

    def __init__(self, *fields, **options):
        """Init panel"""
        super().__init__(**options)
        self.fields = fields


class Table(Component, ComponentFieldsMixin):
    """
    Bootstrap table

    footer: array with first items array/queryset and other items are the fields,
            Same way how the constructor works

    """

    template_name = 'trionyx/components/table.html'

    def __init__(self, objects, *fields, **options):
        """Init Table"""
        footer = options.pop('footer', None)
        super().__init__(**options)

        self.objects = objects
        """Can be string with field name relation, Queryset or list"""

        self.fields = fields

        self.footer_objects = footer[0] if footer else None
        """Can be string with field name relation, Queryset or list"""

        self.footer_fields = footer[1:] if footer else []

    def get_footer_fields(self):
        """Get all footer fields"""
        if not hasattr(self, '__footer_fields'):
            self.__footer_fields = [
                self.parse_field(field, index)
                for index, field in enumerate(self.footer_fields)
            ]
        return self.__footer_fields

    def get_rendered_footer_object(self, obj):
        """Render footer object"""
        return [
            {
                **field,
                'value': self.render_field(field, obj)
            }
            for field in self.get_footer_fields()
        ]

    def get_rendered_footer_objects(self):
        """Render footer objects"""
        objects = self.footer_objects

        if isinstance(objects, str):
            objects = getattr(self.object, objects).all()

        return [
            self.get_rendered_footer_object(obj)
            for obj in objects
        ]



"""
trionyx.trionyx.renderers
~~~~~~~~~~~~~~~~~~~~~~~~~

:copyright: 2018 by Maikel Martens
:license: GPLv3
"""


def render_level(model, *args, **kwargs):
    """Render level as label"""
    mapping = {
        10: 'info',
        20: 'info',
        30: 'warning',
        40: 'danger',
        50: 'danger',
    }

    return '<span class="label label-{}">{}</span>'.format(
        mapping.get(model.level, 'secondary'),
        model.get_level_display().upper()
    )


def render_progress(model, *args, **kwargs):
    """Render progress"""
    from trionyx.trionyx.layouts import progress_renderer
    return progress_renderer(model.progress)


def render_status(model, *args, **kwargs):
    """Render level as label"""
    mapping = {
        10: 'warning',
        50: 'success',
        99: 'danger',
    }

    return '<span class="label label-{}">{}</span>'.format(
        mapping.get(model.status, 'info'),
        model.get_status_display().upper()
    )



"""
trionyx.trionyx.models
~~~~~~~~~~~~~~~~~~~~~~

:copyright: 2018 by Maikel Martens
:license: GPLv3
"""
import hashlib
import traceback
from contextlib import contextmanager

from celery import current_app
from celery.app.control import Control
from django.contrib.auth.models import BaseUserManager, AbstractBaseUser, PermissionsMixin
from django.contrib.contenttypes import fields
from django.utils import timezone
from django.conf import settings
from django.utils import translation
from django.utils.translation import ugettext_lazy as _
from django.core.mail import EmailMultiAlternatives
from django.template.loader import render_to_string
from trionyx import models
from trionyx.utils import get_current_request
from trionyx.data import TIMEZONES


# =============================================================================
# User models
# =============================================================================
class UserManager(BaseUserManager, models.BaseManager):
    """Manager for user"""

    def _create_user(self, email, password, is_superuser, **extra_fields):
        """Create new user"""
        now = timezone.now()
        if not email:
            raise ValueError('The given email must be set')

        email = self.normalize_email(email)
        user = self.model(
            email=email,
            password=password,
            is_active=True,
            is_superuser=is_superuser, last_login=now,
            date_joined=now,
            **extra_fields)
        user.set_password(password)
        user.save(using=self._db)
        return user

    def create_user(self, email, password=None, **extra_fields):
        """Create standard user"""
        return self._create_user(email, password, False, **extra_fields)

    def create_superuser(self, email, password=None, **extra_fields):
        """Create super user"""
        return self._create_user(email, password, True, **extra_fields)

    def get_queryset(self):
        """Get queryset default filter inactive users"""
        return super().get_queryset().filter(is_active=True)


def default_language():
    """Return default language from settings, prevent new migrations if setting changed"""
    return settings.LANGUAGE_CODE


def default_timezone():
    """Return default timezone from settings, prevent new migrations if setting changed"""
    return settings.TIME_ZONE


class User(models.BaseModel, AbstractBaseUser, PermissionsMixin):
    """User model"""

    email = models.EmailField(_('Email'), max_length=255, unique=True)
    first_name = models.CharField(_('First name'), max_length=64, blank=True, default='')
    last_name = models.CharField(_('Last name'), max_length=64, blank=True, default='')
    is_active = models.BooleanField(_('Active'), default=True)
    date_joined = models.DateTimeField(_('Date joined'), default=timezone.now)
    last_online = models.DateTimeField(_('Last online'), blank=None, null=True)
    avatar = models.ImageField(_('Avatar'), blank=True, upload_to='avatars/', default='')

    language = models.CharField(_('Language'), max_length=6, choices=settings.LANGUAGES, default=default_language)
    timezone = models.CharField(_('Timezone'), max_length=32, choices=TIMEZONES, default=default_timezone)

    USERNAME_FIELD = 'email'

    objects = UserManager()

    class Meta:
        """Model meta description"""

        verbose_name = _('User')
        verbose_name_plural = _('Users')

    def get_full_name(self):
        """Get full username if no name is set email is given"""
        if self.first_name and self.last_name:
            return "{} {}".format(self.first_name, self.last_name)
        return self.email

    def get_short_name(self):
        """Get short name if no name is set email is given"""
        if self.first_name:
            return self.first_name
        return self.email

    def set_attribute(self, code, value):
        """Set user attribute"""
        models.get_class(UserAttribute).objects.set_attribute(self, code, value)

    def get_attribute(self, code, default=None):
        """Get user attribute"""
        return models.get_class(UserAttribute).objects.get_attribute(self, code, default)

    @contextmanager
    def locale_override(self):
        """Override locale settings to user settings"""
        with translation.override(self.language), timezone.override(self.timezone):
            yield

    def send_email(self, subject, body='', html_template=None, template_context=None, files=None):
        """Send email to user"""
        if not body and not html_template:
            raise Exception('You must supply a body or/and html_template')

        with self.locale_override():
            message = EmailMultiAlternatives(
                subject=subject,
                body=body,
                from_email=settings.DEFAULT_FROM_EMAIL,
                to=[self.email]
            )

            if html_template:
                message.attach_alternative(
                    render_to_string(html_template, template_context if template_context else {}),
                    "text/html")

        if files:
            for file in files:
                message.attach(file.name, file.read())

        return message.send()


class UserAttributeManager(models.Manager):
    """User attribute manager"""

    def set_attribute(self, user, code, value):
        """Set attribute for user"""
        self.update_or_create(user=user, code=code, defaults={'value': value})

    def get_attribute(self, user, code, default=None):
        """Get attribute for user"""
        try:
            return self.get(user=user, code=code).value
        except models.ObjectDoesNotExist:
            return default


class UserAttribute(models.Model):
    """User attribute to store system values for user"""

    user = models.ForeignKey(models.get_name(User), models.CASCADE, related_name='attributes')
    code = models.CharField(max_length=128, null=False)
    value = models.JSONField()

    objects = UserAttributeManager()

    class Meta:
        """Model meta description"""

        unique_together = ('user', 'code')

    def __str__(self):
        """User Attribute representation"""
        return self.code


# =============================================================================
# Logging
# =============================================================================
class LogManager(models.BaseManager):
    """Log manager"""

    def create_log_entry_by_record(self, record):
        """Create log entry by `logging.LogRecord`"""
        log_hash = hashlib.md5(str(' '.join(str(x) for x in [
            record.pathname,
            record.lineno,
            record.msg,
        ])).encode()).hexdigest()

        log, _ = self.get_or_create(log_hash=log_hash, defaults={
            'level': record.levelno,
            'message': record.msg,
            'file_path': record.pathname,
            'file_line': record.lineno,
            'last_event': timezone.now(),
        })

        if record.exc_info and not log.traceback:
            log.traceback = ''.join(traceback.TracebackException(
                *record.exc_info
            ).format())

        request = get_current_request()
        entry = LogEntry.objects.create(
            log=log,
            log_time=timezone.now(),
            user=request.user if request and not request.user.is_anonymous else None,
            path=request.path if request else '',
            user_agent=request.META.get('HTTP_USER_AGENT') if request else '',
        )

        log.last_event = entry.log_time
        log.log_count = log.entries.count()
        log.save()


class Log(models.BaseModel):
    """Log"""

    CRITICAL = 50
    ERROR = 40
    WARNING = 30
    INFO = 20
    DEBUG = 10
    NOTSET = 0

    LEVEL_CHOICES = [
        (CRITICAL, _('Critical')),
        (ERROR, _('Error')),
        (WARNING, _('Warning')),
        (INFO, _('Info')),
        (DEBUG, _('Debug')),
        (NOTSET, _('Not set')),
    ]

    log_hash = models.CharField(_('Log hash'), max_length=32)
    level = models.IntegerField(_('Level'), choices=LEVEL_CHOICES)
    message = models.TextField(_('Message'))
    file_path = models.CharField(_('File path'), max_length=256)
    file_line = models.IntegerField(_('File line'))
    traceback = models.TextField(_('Traceback'), default='')

    last_event = models.DateTimeField(_('Last event'))
    log_count = models.IntegerField(_('Log count'), default=1)

    objects = LogManager()

    class Meta:
        """Model meta description"""

        verbose_name = _('Log')
        verbose_name_plural = _('Logs')


class LogEntry(models.Model):
    """Log entry event"""

    log = models.ForeignKey(Log, models.CASCADE, related_name='entries')
    log_time = models.DateTimeField(_('Log time'))

    user = models.ForeignKey(models.get_name(User), models.SET_NULL, null=True, blank=True, verbose_name=_('User'))
    path = models.TextField(_('Path'), default='')
    user_agent = models.TextField(_('User agent'), default='')

    class Meta:
        """Model meta description"""

        verbose_name = _('Log entry')
        verbose_name_plural = _('Log entries')


class AuditLogEntry(models.BaseModel):
    """Auditlog model"""

    ACTION_ADDED = 10
    ACTION_CHANGED = 20
    ACTION_DELETED = 30

    action_choices = [
        (ACTION_ADDED, _('Added')),
        (ACTION_CHANGED, _('Changed')),
        (ACTION_DELETED, _('Deleted')),
    ]

    content_type = models.ForeignKey('contenttypes.ContentType', models.CASCADE, related_name='+')
    object_id = models.BigIntegerField(blank=True, null=True)
    content_object = fields.GenericForeignKey('content_type', 'object_id')
    object_verbose_name = models.TextField(default='', blank=True)

    user = models.ForeignKey(models.get_name(User), models.SET_NULL, blank=True, null=True, related_name='+')
    action = models.IntegerField(choices=action_choices)
    changes = models.JSONField()

    class Meta:
        """Model meta description"""

        indexes = [
            models.Index(fields=['content_type', 'object_id']),
        ]


class Task(models.BaseModel):
    """Task model"""

    SCHEDULED = 10
    QUEUE = 20
    LOCKED = 30
    RUNNING = 40
    COMPLETE = 50
    FAILED = 99

    STATUS_CHOICES = (
        (SCHEDULED, _('Scheduled')),
        (QUEUE, _('Queue')),
        (LOCKED, _('Locked')),
        (RUNNING, _('Running')),
        (COMPLETE, _('Completed')),
        (FAILED, _('Failed')),
    )

    celery_task_id = models.CharField(max_length=64, unique=True)
    status = models.IntegerField(_('Status'), choices=STATUS_CHOICES, default=QUEUE)
    identifier = models.CharField(_('Identifier'), max_length=255, default='not_set')
    description = models.TextField(_('Description'), default='')

    progress = models.PositiveIntegerField(_('Progress'), default=0)
    progress_output = models.JSONField(_('Progress output'), default=list)

    scheduled_at = models.DateTimeField(_('Scheduled at'), blank=True, null=True)
    started_at = models.DateTimeField(_('started at'), blank=True, null=True)
    execution_time = models.IntegerField(_('Execution time'), default=0)
    result = models.TextField(_('Result'), blank=True, default='')

    user = models.ForeignKey(
        models.get_name(User), models.SET_NULL, blank=True, null=True,
        verbose_name=_('Started by'))
    object_type = models.ForeignKey(
        'contenttypes.ContentType',
        models.SET_NULL,
        blank=True,
        null=True,
        related_name='+',
        verbose_name=_('Object type')
    )
    object_id = models.BigIntegerField(_('Object ID'), blank=True, null=True)
    object = fields.GenericForeignKey('object_type', 'object_id')
    object_verbose_name = models.TextField(_('Object name'), blank=True, default='')

    class Meta:
        """Model meta description"""

        verbose_name = _('Task')
        verbose_name_plural = _('Tasks')

    def cancel_celery_task(self, kill=False):
        """
        Make sure we cancel the task (if in queue/scheduled).
        :param: kill Also kill the task if it's running, defaults to False.
        """
        celery_control = Control(current_app)
        celery_control.revoke(task_id=self.celery_task_id, terminate=kill)



"""
trionyx.trionyx.cron
~~~~~~~~~~~~~~~~~~~~

:copyright: 2019 by Maikel Martens
:license: GPLv3
"""
from datetime import timedelta


schedule = {
    'cleanup_unexpectedly_stopped_tasks': {
        'task': 'trionyx.trionyx.tasks.cleanup_unexpectedly_stopped_tasks',
        'schedule': timedelta(minutes=15)
    }
}



"""
trionyx.navigation
~~~~~~~~~~~~~~~~~~

:copyright: 2017 by Maikel Martens
:license: GPLv3
"""
import re
import copy

from django.apps import apps
from django.urls import reverse

from trionyx.config import models_config


class Menu:
    """Meu class that hold the root tree item"""

    def __init__(self, root_item=None):
        """Init Menu"""
        self.root_item = root_item

    def auto_load_model_menu(self):
        """
        Auto load model menu entries, can be configured in `trionyx.config.ModelConfig`:

        - menu_name
        - menu_icon
        - menu_order
        """
        from trionyx.trionyx.apps import BaseConfig

        order = 0
        for app in apps.get_app_configs():
            if not isinstance(app, BaseConfig) or getattr(app, 'no_menu', False):
                continue

            app_path = app.name.split('.')[-1]
            model_order = 0
            for model in app.get_models():
                config = models_config.get_config(model)
                if config.menu_exclude:
                    continue

                menu_icon = None
                menu_path = '{}/{}'.format(app_path, config.model_name)
                if config.menu_root:
                    order += 10
                    menu_order = order
                    menu_icon = config.menu_icon
                    menu_path = config.model_name
                else:
                    model_order += 10
                    menu_order = model_order

                self.add_item(
                    path=menu_path,
                    name=config.menu_name if config.menu_name else model._meta.verbose_name_plural.capitalize(),
                    order=config.menu_order if config.menu_order else menu_order,
                    icon=menu_icon,
                    url=reverse(
                        "trionyx:model-list",
                        kwargs={
                            'app': model._meta.app_label,
                            'model': model._meta.model_name,
                        }
                    ),
                    permission='{app_label}.view_{model_name}'.format(
                        app_label=config.app_label,
                        model_name=config.model_name,
                    ).lower()
                )

            if model_order > 0:
                order += 10
                self.add_item(
                    path=app_path,
                    name=getattr(app, 'menu_name', app.verbose_name),
                    icon=getattr(app, 'menu_icon', None),
                    order=getattr(app, 'menu_order', order),
                )

    def add_item(self, path, name, icon=None, url=None, order=None, permission=None, active_regex=None):
        """
        Add new menu item to menu

        :param path: Path of menu
        :param name: Display name
        :param icon: CSS icon
        :param url: link to page
        :param order: Sort order
        :param permission:
        :return:
        """
        if self.root_item is None:
            self.root_item = MenuItem('ROOT', 'ROOT')

        root_item = self.root_item
        current_path = ''
        for node in path.split('/')[:-1]:
            if not node:
                continue

            current_path = '/' + '{}/{}'.format(current_path, node).strip('/')
            new_root = root_item.child_by_code(node)
            if not new_root:  # Create menu item if not exists
                new_root = MenuItem(current_path, name=str(node).capitalize())
                root_item.add_child(new_root)
            root_item = new_root

        new_item = MenuItem(path, name, icon, url, order, permission, active_regex)

        current_item = root_item.child_by_code(path.split('/')[-1])
        if current_item:
            current_item.merge(new_item)
        else:
            root_item.add_child(new_item)

    def get_menu_items(self, user=None):
        """Get menu items"""
        if not self.root_item:
            return []

        def filter_childs(childs):
            menu = []
            for item in childs:
                if not item.permission or (item.permission and user.has_perm(item.permission)):
                    item = copy.copy(item)
                    if item.childs:
                        item.childs = filter_childs(item.childs)
                        if item.childs:
                            menu.append(item)
                    else:
                        menu.append(item)

            return menu

        return filter_childs(self.root_item.childs)


class MenuItem:
    """Menu item that holds all menu item data"""

    def __init__(self, path, name, icon=None, url=None, order=None, permission=None, active_regex=None):
        """
        Init menu item

        :param path: Path of menu
        :param name: Display name
        :param icon: CSS icon
        :param url: link to page
        :param order: Sort order
        :param permission:
        """
        self.path = path
        self.name = name
        self.icon = icon
        self.url = url
        self.order = order
        self.permission = permission
        self.active_regex = active_regex

        self.depth = 0
        self.childs = []

    def merge(self, item):
        """Merge Menu item data"""
        self.name = item.name

        if item.icon:
            self.icon = item.icon

        if item.url:
            self.url = item.url

        if item.order:
            self.order = item.order

        if item.permission:
            self.permission = item.permission

    def add_child(self, item):
        """Add child to menu item"""
        item.depth = self.depth + 1
        self.childs.append(item)
        self.childs = sorted(self.childs, key=lambda item: item.order if item.order else 999)

    def child_by_code(self, code):
        """
        Get child MenuItem by its last path code

        :param code:
        :return: MenuItem or None
        """
        for child in self.childs:
            if child.path.split('/')[-1] == code:
                return child
        return None

    def is_active(self, path):
        """Check if given path is active for current item"""
        if self.url == '/' and self.url == path:
            return True
        elif self.url == '/':
            return False

        if self.url and path.startswith(self.url):
            return True

        if self.active_regex and re.compile(self.active_regex).match(path):
            return True

        for child in self.childs:
            if child.is_active(path):
                return True
        return False


app_menu = Menu()
add_menu_item = app_menu.add_item



from trionyx.trionyx.apps import BaseConfig


class BlogConfig(BaseConfig):
    """Django core config app"""

    name = 'app.testblog'
    verbose_name = 'Blog'

    menu_name = 'Super blog'
    menu_icon = 'fa fa-rss-square'
    menu_order = 500

    class Post:
        verbose_name = '{title}'
        search_description = '{publish_date} {title} {price}'
        list_default_fields = ['id', 'publish_date', 'title']

        view_header_buttons = [
            {
                'label': 'Publish',  # string or function
                'url': 'trionyx:model-dialog-edit',  # string or function
                'type': 'bg-theme', # string or function
                'show': lambda obj, alias: True,  # Function that gives True or False if button must be displayed
                'modal': True,
            }
        ]

    class Category:
        verbose_name = '{name}'

        list_fields = [
            {
                'field': 'name',
                'renderer': lambda model, field: model.name.upper()
            }
        ]

        list_default_fields = ['id', 'created_at', 'name']

    class Tag:
        verbose_name = '{name}'
        menu_exclude = True